from math import pi, sin, cos

from direct.showbase.ShowBase import ShowBase
from panda3d.core import loadPrcFile
from panda3d.core import DirectionalLight, AmbientLight
from panda3d.core import TransparencyAttrib
from panda3d.core import WindowProperties
from panda3d.core import CollisionTraverser, CollisionNode, CollisionBox, CollisionRay, CollisionHandlerQueue
from panda3d.core import LineSegs, PandaNode, NodePath
from direct.gui.OnscreenImage import OnscreenImage

loadPrcFile('settings.prc')

def degToRad(degrees):
    return degrees * (pi / 180.0)

MOUSE_ROTATE_SPEED = 12.0
CAMERA_MOVE_SPEED = 8.0

class MyGame(ShowBase):
    def __init__(self):
        ShowBase.__init__(self)

        self.loadModels()
        self.setupLights()
        self.generateWorld()
        self.makeBlock()
        self.setupCamera()
        # self.setupSkybox()
        #self.captureMouse()
        self.setupControls()
        self.mouseDown = False
        taskMgr.add(self.update, 'update')


    def setupControls(self):
        self.accept('mouse1', lambda: self.handleClick("left"))
        self.accept('mouse1-up', self.clearMouse)
        self.accept('mouse3', lambda: self.handleClick("right"))
        self.accept('mouse3-up', self.clearMouse)


# ---------------------------  Mouse picking and control  ------------------------       

    def handleClick(self, which):
        self.mouseDown = which
        md = self.win.getPointer(0)
        mx, my = md.getX(), md.getY()
        if md.getInWindow():
            self.lastMouseX = mx
            self.lastMouseY = my
            what, status = self.pickSomething(mx, my)
            if status == "played":
                self.mouseDown = "right" # a dirty hack; rotating the world
            elif status == "active":
                self.mouseDown = "left" # mirror image of the same dirty hack

    def pickSomething(self, mouseX, mouseY):
        # self.pickerRay.setFromLens(self.camNode, mouseX, mouseY)
        # self.cTrav.traverse(render)
        if self.rayQueue.getNumEntries() > 0:
            print("picking hit something")
            self.rayQueue.sortEntries()
            rayHit = self.rayQueue.getEntry(0)
            hitNodePath = rayHit.getIntoNodePath()
            ancestors = hitNodePath.getAncestors()
            hit = hitNodePath.findNetTag("status")
            if hit:
                status = hit.getTag("status")
                return hit, status
        return None, "played"

    def clearMouse(self):
        self.mouseDown = False
    

    # ---------------------------  User viewing control  ------------------------       

    # A GraphicsWindow object contains getPointer
    # this returns a PointerData object
    def update(self, task):  # to move block
        dt = globalClock.getDt()

        # *** mouse movement turns the object
        if self.mouseDown:
            md = self.win.getPointer(0)
            mouseX = md.getX()
            mouseY = md.getY()
            mouseChangeX = mouseX - self.lastMouseX
            mouseChangeY = mouseY - self.lastMouseY
            
            if self.mouseDown == "left":
                self.myBlock.setHpr(self.myBlock,
                     mouseChangeX * dt * MOUSE_ROTATE_SPEED,
                    min(90, max(-90,  mouseChangeY * dt * MOUSE_ROTATE_SPEED)),
                    0
                )
            elif self.mouseDown == "right":
                self.world.setHpr(self.world,
                     mouseChangeX * dt * MOUSE_ROTATE_SPEED,
                    min(90, max(-90,  mouseChangeY * dt * MOUSE_ROTATE_SPEED)),
                    0
                )            
            self.lastMouseX = mouseX
            self.lastMouseY = mouseY

        return task.cont

# ---------------------------  Graphics machine stuff  ------------------------       

    def setupLights(self):
        mainLight = DirectionalLight('main light')
        mainLightNodePath = render.attachNewNode(mainLight)
        mainLightNodePath.setHpr(30, -60, 0)
        render.setLight(mainLightNodePath)

        ambientLight = AmbientLight('ambient light')
        ambientLight.setColor((0.3, 0.3, 0.3, 1))
        ambientLightNodePath = render.attachNewNode(ambientLight)
        render.setLight(ambientLightNodePath)    
        
        
    def setupCamera(self):
        self.disableMouse()
        self.camera.setPos(0, -10, 3)
        self.camLens.setFov(80)

        crosshairs = OnscreenImage(
            image = 'crosshairs.png',
            pos = (0, 0, 0),
            scale = 0.05,
        )
        crosshairs.setTransparency(TransparencyAttrib.MAlpha)

        self.cTrav = CollisionTraverser()
        ray = CollisionRay()
        ray.setFromLens(self.camNode, (0, 0))
        rayNode = CollisionNode('line-of-sight')
        rayNode.addSolid(ray)
        self.pickerRay = ray
        rayNodePath = self.camera.attachNewNode(rayNode)
        self.rayQueue = CollisionHandlerQueue()
        self.cTrav.addCollider(rayNodePath, self.rayQueue)

# -----------------------------   Build the world  -------------------------------
    def generateWorld(self):
        self.world = render.attachNewNode("world")
        for z in range(10):
            for y in range(20):
                for x in range(20):
                    self.createNewBlock(
                        x * 2 - 20,
                        y * 2 - 20,
                        -z * 2,
                        'stone'
                    )


    def createNewBlock(self, x, y, z, type):
        newBlockNode = self.world.attachNewNode('new-block-placeholder')
        newBlockNode.setPos(x, y, z)

        self.stoneBlock.instanceTo(newBlockNode)

        blockSolid = CollisionBox((-1, -1, -1), (1, 1, 1))
        blockNode = CollisionNode('block-collision-node')
        blockNode.addSolid(blockSolid)
        collider = newBlockNode.attachNewNode(blockNode)
        collider.setPythonTag('owner', newBlockNode)

    def loadModels(self):
        global stony
        self.stoneBlock = loader.loadModel('stone-block.glb')
        stony = self.stoneBlock


# build a block
    def makeBlock(self):
        mainBlock = self.world.attachNewNode("main-block")
        mainBlock.setTag("status", "active")
        block1 = mainBlock.attachNewNode("block1")
        block2 = mainBlock.attachNewNode("block2")
        block3 = mainBlock.attachNewNode("block3")
        block2.setPos(2, 0, 0)
        block3.setPos(2, 0, 2)
        stony.instanceTo(block1)
        stony.instanceTo(block2)
        stony.instanceTo(block3)
        self.myBlock = mainBlock
        self.myBlock.setPos(0, 8, 4)


game = MyGame()
game.run()
